/*
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Created on 5 dicembre 2006, 17.12
 *
 */

package com.funambol.mailclient.mm;

import com.funambol.mail.Message;
import com.funambol.mail.MessageFlags;
import com.funambol.mailclient.AccountListener;
import com.funambol.mailclient.cm.ContactManager;
import com.funambol.mailclient.cm.ContactManagerException;
import com.funambol.mailclient.cm.ContactManagerFactory;
import com.funambol.mailclient.config.ConfigException;
import com.funambol.mailclient.config.ConfigManager;
import com.funambol.mailclient.loc.Localization;
import com.funambol.syncml.spds.SyncException;
import com.funambol.util.MailDateFormatter;


import com.funambol.util.Log;
import com.funambol.util.Observer;
import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.Hashtable;
import javax.microedition.rms.RecordStoreException;

/**
 * Extends TaskManager to monitor a sync thread.
 * <p>
 * update() method of observer is called when something changes.
 * The object passed to the observer gives a hint of what has changed.
 * </p>
 * <ul>
 * <li>each time the userInfoMessage is changed an update(:String) is sent,
 *     and the string passed contains the new userInfoMessage</li>
 * <li>each time a new userInfoMessage is received an update(:Message) is
 *     sent. Message is an empty userInfoMessage</li>
 * </ul>
 */
public class SyncTaskManager extends TaskManager {
    private static final int MAX = 100;
    
    private int toReceive = AccountListener.MESSAGE_NUMBER_UNKNOWN;
    private int toSend = AccountListener.MESSAGE_NUMBER_UNKNOWN;
    private int received = 0;
    private int sent = 0;
    
    private boolean connected = false;
    private boolean completed = false;
    
    
    // FIXME: this should be called syncContacts!
    private boolean resetContacts = true;
    
    private MessageManager mm = null;
    
    // SyncEventInfo permits to propagate to the upper level the info for
    // each Sync Event
    private SyncEventInfo syncEventInfo;
    
    private int syncMode;
    
    private boolean syncMessages;
    
    /**
     * Creates a new instance of SyncTaskManager
     *
     * @param MessageManager the userInfoMessage manager
     * @numObservers the max number of observers this
     * task manager will be able to have
     */
    public SyncTaskManager(MessageManager mm, int numObservers) {
        super(numObservers);
        this.mm = mm;
        mm.setListener(this);
        max = MAX;
        syncEventInfo = new SyncEventInfo();
        this.syncMode = -1;
    }
    
    /**
     * Returns the current sync mode, -1 means that the one defined for the
     * source will be used.
     */
    public int getSyncMode() {
        return syncMode;
    }
    
    /**
     * Force to use a sync mode different from the one defined for the source.
     */
    public void setSyncMode(int mode) {
        syncMode = mode;
    }
    
    /**
     * Javabean standard definition for 'syncContacts' field setter method
     *
     * @param resetContacts is the new value to give to resetContacts variable
     */
    public void setResetContacts(boolean syncContacts) {
        this.resetContacts = syncContacts;
    }
    
    /**
     * Javabean standard definition for 'resetContacts' field setter method
     *
     * @return boolean value related to resetContacts variable
     */
    public boolean isResetContacts() {
        return resetContacts;
    }
    
    /**
     * @return true is task is finished
     */
    public boolean isFinished() {
        return finished;
    }
    
    /**
     * run the sync
     */
    public void run() {
        //Sync Mail Messages;
        
        if (syncMessages) {
            
            syncEventInfo.setInfoCode(SyncEventInfo.START_OPERATION);
            update(syncEventInfo);
            try {
                Log.info("*** Initial Sync for Emails in progress ***");
                
                //setting message to have better response time
                syncEventInfo.setInfoCode(SyncEventInfo.START_MESSAGE_SESSION);
                syncEventInfo.infoMessage = Localization.getMessages().CONNECTING;
                update(syncEventInfo);
                
                // Bad trick to use the source mode. XXX
                if(syncMode == -1) {
                    mm.sync();
                } else {
                    mm.sync(syncMode);
                }
                
                Log.info("*** Initial Sync for Emails is complete ***");
                
            } catch (ConfigException ce) {
                Log.error(this, "Config error" + ce.toString());
                userInfoMessage = Localization.getMessages().CONFIGURATION_ERROR;
                
            } finally {
                syncEventInfo.setInfoCode(SyncEventInfo.END_MESSAGE_SESSION);
                update(syncEventInfo);
                
            }
            finished = true;
        }
        
        //Sync Email Client Address Book;
        //check resetContacts attribute: if it's set to true
        //a contact Synchronization Takes place after message sync.
        //This is needed when the application is started
        //or a new account with a different username is
        //declared and saved into configuration by the user;
        if (this.resetContacts) {
            try {
                Log.info("*** Initial Sync for Contacts in progress ***");
                syncEventInfo.setInfoCode(SyncEventInfo.START_CONTACT_SESSION);
                syncEventInfo.infoMessage =
                        Localization.getMessages().SYNC_CONTACTS;
                update(syncEventInfo);
                ContactManager cm =
                        ContactManagerFactory.getContactManager(ContactManager.RMS_TYPE);
                // Bad trick to use the source mode. XXX
                if(syncMode == -1) {
                    cm.sync();
                } else {
                    cm.sync(syncMode);
                }
                Log.info("*** Initial Sync for Contacts is complete ***");
                
                if (syncMessages) {
                    userInfoMessage = Localization.getMessages().getX_MESSAGES_RECEIVED(received);
                } else   {
                    userInfoMessage = Localization.getMessages().SYNC_CONTACTS_COMPLETED;
                }
            } catch (RecordStoreException ex) {
                Log.error("Error updating contacts:"+ex.toString());
                ex.printStackTrace();
            } catch (ContactManagerException ex) {
                Log.error("Error updating contacts:"+ex.toString());
                ex.printStackTrace();
            } catch (ConfigException ex) {
                Log.error(this, "Config error" + ex.toString());
                userInfoMessage = Localization.getMessages().CONFIGURATION_ERROR;
            } finally {
                this.resetContacts=false;
                syncEventInfo.infoMessage = userInfoMessage;
                syncEventInfo.setInfoCode(SyncEventInfo.END_CONTACT_SESSION);
                update(syncEventInfo);
                mm.setListener(null);
            }
        }
        
        syncEventInfo.setInfoCode(SyncEventInfo.END_OPERATION);
        update(syncEventInfo);
    }
    
    /**
     * should be called when a communication session is started
     */
    public void startSession() {
        syncEventInfo.setInfoCode(SyncEventInfo.START_MESSAGE_SESSION);
        syncEventInfo.infoMessage =
                Localization.getMessages().CONNECTING;
        update(syncEventInfo);
    }
    
    /**
     * should be called when connnection is completed
     */
    public void connectionCompleted(int onConnectAction) {
        connected = true;
        if (onConnectAction == AccountListener.ACTION_RESET_INBOX) {
            syncEventInfo.setInfoCode(SyncEventInfo.RELOAD_INBOX);
            update(syncEventInfo);
        }
        if (toSend == 0 || toSend == MESSAGE_NUMBER_UNKNOWN ) {
            syncEventInfo.setInfoCode(SyncEventInfo.CONNECTION_COMPLETED);
            syncEventInfo.infoMessage =
                    Localization.getMessages().LOADING;
            update(syncEventInfo);
        }
    }
    
    /**
     * Invoked each time data is received from the server, with the timestamp
     * and the size of the received data.
     */
    public void dataReceived(String date, int size) {
        
        // date like Thu, 03 May 2007 14:45:38 GMT
        try{
            if (!ConfigManager.getConfig().getTimeZoneUpdated()) {
                
                Date gmtServerdate = MailDateFormatter.parseRfc2822Date(date);
                Date deviceDate = new Date();
                
                int millisServerDate = (int)gmtServerdate.getTime();
                int millisDeviceDate = (int)deviceDate.getTime();
                int hourServer = (int)(millisServerDate/3600000); //60*60*1000
                int hourDevice = (int)(millisDeviceDate/3600000); //60*60*1000
                int hourDiff = hourDevice - hourServer;
                int minServer = (int)(millisServerDate/60000); //60*60*1000
                int minDevice = (int)(millisDeviceDate/60000); //60*60*1000
                int minDiff = minDevice - minServer - (hourDiff*60);
                StringBuffer offset = new StringBuffer();
                offset.append((hourDiff > 0) ? "+" : "-");
                int mathHourDiff = (Math.abs(hourDiff));
                offset.append((mathHourDiff<=9) ? "0"+mathHourDiff : ""+mathHourDiff);
                int mathMinDiff = (Math.abs(minDiff));
                if(mathMinDiff <= 20 || mathMinDiff >=55)
                    offset.append("00");
                else if(mathMinDiff > 20 || mathMinDiff < 40  )
                    offset.append("30");
                else if (mathMinDiff >=40 || mathMinDiff < 55)
                    offset.append("45");
                
                // Save the new config.
                ConfigManager.getConfig().setTimeZone(offset.toString());
                ConfigManager.saveConfig();
            }
        } catch(ConfigException cex){
            Log.error(this, "dataReceived - " + cex.toString());
            userInfoMessage = Localization.getMessages().CONFIGURATION_ERROR;
        } catch(IOException ioex){
            Log.error(this, "dataReceived - " + ioex.toString());
        }
    }
    
    /**
     * should be called when a userInfoMessage is received
     */
    public void msgReceived(Message msg) {
        this.received++;
        update(msg);
        
        if (toReceive != AccountListener.MESSAGE_NUMBER_UNKNOWN) {
            userInfoMessage = Localization.getMessages().
                    getMESSAGE_X_OF_Y_RECEIVED(received,toReceive);
        } else {
            userInfoMessage = Localization.getMessages().
                    getMESSAGE_X_RECEIVED(received);
        }
        
        syncEventInfo.setInfoCode(SyncEventInfo.MESSAGE_RECEIVED);
        syncEventInfo.infoMessage = userInfoMessage;
        update(syncEventInfo);
    }
    
    /**
     * should be called when a userInfoMessage is sent
     */
    public void msgSent() {
        sent++;
        /*
        if (toSend != AccountListener.MESSAGE_NUMBER_UNKNOWN) {
            userInfoMessage = Localization.getMessages("MessageXofYSent",
                    new Object[]{new Integer(sent), new Integer(this.toSend)});
        } else {
            userInfoMessage = Localization.getMessages("MessageXSent",
                    new Object[]{new Integer(sent)});
        }
        syncEventInfo.setInfoCode(SyncEventInfo.MESSAGE_SENT);
        syncEventInfo.infoMessage = userInfoMessage;
        update(syncEventInfo);
         */
    }
    
    
    /**
     * should be called when a message session ends
     *@param status the status as returned by the server
     *
     */
    public void endSession(int status) {
        
        if (status != AccountListener.STATUS_SUCCESS) {
            userInfoMessage = getErrorMessage(status);
        }  else {
            userInfoMessage = Localization.getMessages().getX_MESSAGES_RECEIVED(received);
        }

        syncEventInfo.setInfoCode(SyncEventInfo.END_MESSAGE_SESSION);
        syncEventInfo.infoMessage = userInfoMessage;
        update(syncEventInfo);
    }
    
    /**
     * Should be called when receiving is started
     *
     *
     * @param number The number of userInfoMessage to be received, or
     *               AccountListener.MESSAGE_NUMBER_UNKNOWN if unknown
     */
    public void startReceiving(int number) {
        toReceive = number;
        /*
        if (toReceive != AccountListener.MESSAGE_NUMBER_UNKNOWN) {
            userInfoMessage = Localization.getMessages("incoming_messages",
                    new Object[]{new Integer(number)});
        }
        syncEventInfo.setInfoCode(SyncEventInfo.START_RECEIVE);
        syncEventInfo.infoMessage = userInfoMessage;
        update(syncEventInfo);
         */
    }
    
    /**
     * should be called when started sending.
     *
     * @param the number of userInfoMessage to be sent,
     * or AccountListener.MESSAGE_NUMBER_UNKNOWN if
     * unknown
     */
    public void startSending(int number) {
        Log.debug(this, " #########################    Start sending " + number);
        
        toSend = number;
        
        //if we have something to send
        if (toSend != 0) {// && toSend != AccountListener.MESSAGE_NUMBER_UNKNOWN) {
            if (toSend == AccountListener.MESSAGE_NUMBER_UNKNOWN) {
                userInfoMessage = Localization.getMessages().SENDING_DOTTED;
            } else if (toSend > 0) {
                userInfoMessage = Localization.getMessages().getSENDINGX(toSend);
            }
            syncEventInfo.setInfoCode(SyncEventInfo.START_SEND);
            syncEventInfo.infoMessage = userInfoMessage;
            update(syncEventInfo);
        }
        
        
        
    }
    
    /**
     * should be called when finished sending
     */
    public void endSending() {
        //TODO
    }
    
    /*
     * calculate and update progress
     *
    private void updateProgress() {
        progress = 0;
     
        if (connected)
            progress += MAX/4;
     
        // if we've sent at least one item
        if (sent > 0 || toSend >= 0)
            if (sent < toSend)
                //if we know the % of sent userInfoMessage we use it to
                //set the progress
                progress += (sent/toSend) * MAX/4;
            else //we increment the progress by 25%
                progress += MAX/4;
        else
            if (received > 0)
                progress += MAX/4;
     
        if (received>0||toReceive>=0)
            if (received < toReceive)
                progress += (received/toReceive) * MAX/4;
            else
                progress += MAX/4;
     
        if (completed)
            progress = MAX;
    }
     */
    
    public void msgDeleted(String msgId) {
        Log.debug("delete notification from server for message " + msgId);
        syncEventInfo.setInfoCode(syncEventInfo.MESSAGE_DELETED);
        syncEventInfo.msgId = msgId;
        update(syncEventInfo);
    }
    
    public void msgFlagsUpdated(String msgId, MessageFlags newFlags) {
        Log.debug("update notification from server for message " + msgId);
        syncEventInfo.setInfoCode(syncEventInfo.MESSAGE_FLAGS_UPDATED);
        syncEventInfo.msgId = msgId;
        syncEventInfo.flags = newFlags;
        update(syncEventInfo);
    }
    
    public void reset() {
        int prevNumObservers = numObservers;
        super.reset();
        observer = new Observer[prevNumObservers];
        toReceive = AccountListener.MESSAGE_NUMBER_UNKNOWN;
        toSend = AccountListener.MESSAGE_NUMBER_UNKNOWN;
        received = 0;
        sent = 0;
        connected = false;
        completed = false;
        resetContacts = true;
        max = MAX;
        syncEventInfo = new SyncEventInfo();
        mm.setListener(this);
        this.syncMode = -1;
    }
    
    
    public boolean isSyncMessages() {
        return syncMessages;
    }
    
    /**
     * choose if the run will perform a sync of the messages
     *
     */
    public void setSyncMessages(boolean syncInbox) {
        this.syncMessages = syncInbox;
    }
    
    
}
