/*
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.funambol.mailclient.syncml;

import com.funambol.mailclient.mm.MessageManager;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Date;

import com.funambol.mailclient.AccountListener;

import com.funambol.syncml.client.BaseSyncSource;
import com.funambol.syncml.protocol.SyncML;
import com.funambol.syncml.protocol.SyncMLStatus;
import com.funambol.syncml.spds.SourceConfig;
import com.funambol.syncml.spds.SyncItem;
import com.funambol.syncml.spds.SyncException;

import com.funambol.mail.Folder;
import com.funambol.mail.Message;
import com.funambol.mail.MessageFlags;
import com.funambol.mail.Store;
import com.funambol.mail.StoreFactory;
import com.funambol.mail.MailException;

import com.funambol.util.Log;
import com.funambol.util.XmlException;

/**
 * An implementation of the <i>SyncSource</i> interface, used to send
 * and receive email over the SyncML protocol.
 *
 */
public class MailSyncSource extends BaseSyncSource {
    
    public static final String FOLDER_TYPE = "application/vnd.omads-folder+xml";
    public static final String EMAIL_TYPE = "application/vnd.omads-email+xml";
    
    //--------------------------------------------------------------- Attributes
    
    /** Reference to the mail store */
    private Store store;
    /** Reference to the account listener */
    private AccountListener listener;
    /** A global status code for the MailSyncSource */
    private int mailSourceStatus;
    /** Flag used to mark the beginning of the receiving phase */
    private boolean receiving;
    
    //------------------------------------------------------------- Constructors
    
    /**
     * MailSyncSource constructor: initialize source config and
     * init all the rest to null. The real initialization is done by
     * the beginSync method.
     */
    public MailSyncSource(SourceConfig config) {
        super(config);
        
        store = StoreFactory.getStore();
        listener = null;
        mailSourceStatus = AccountListener.STATUS_SUCCESS;
        receiving = false;
    }
    
    //----------------------------------------------------------- Public Methods
    
    /**
     * Set an AccountListener for this Mail source.
     */
    public void setListener(AccountListener listener) {
        this.listener = listener;
    }
    
    /**
     * Called after SyncManager preparation and initialization just before start
     * the synchronization of the SyncSource.
     *
     * @param syncMode the synchronization type: one of the values in
     *                 sync4j.framework.core.AlertCode
     *
     * @throws SyncException in case of error. This will stop the sync process
     */
    public void beginSync(int syncMode) throws SyncException {
        super.beginSync(syncMode);      // call BaseSyncSource.beginSync()
         
        
        int action = AccountListener.ACTION_NONE;
        
        if(syncMode == SyncML.ALERT_CODE_REFRESH_FROM_SERVER ||
                syncMode == SyncML.ALERT_CODE_SLOW) {
            
            // Remove local mail in INBOX
            try {
                store.removeFolder(store.INBOX);
                store.createFolder(store.INBOX);
                
                // we need to reset the inbox but we don't want the user
                // to catch us performing in a slow sync required
                // from the server... so we refresh the  inbox 
                // only when we are doing a refresh from server (that
                // has probably been requested by our client)
                
                if ( syncMode == SyncML.ALERT_CODE_REFRESH_FROM_SERVER ) {
                     action  = AccountListener.ACTION_RESET_INBOX;
                }
                
            } catch(MailException se) {
                Log.error("[MailSyncSource.beginSync]:"+se.toString());
                throw new SyncException(
                            SyncException.CLIENT_ERROR,
                            "Error cleaning local mail: " +se.toString());
            }
        }
        
        mailSourceStatus = AccountListener.STATUS_SUCCESS;
        receiving = false;
        
        if(listener != null) {
            listener.connectionCompleted(action);
        }
        
    }
    
    /**
     * Called just before committing the synchronization process by the
     * SyncManager. The SyncSource can stop the commit phase raising an
     * exception here.
     *
     * @throws SyncException in case of error, to stop the commit.
     */
    public void endSync() throws SyncException  {
        super.endSync();        // call BaseSyncSource.endSync()   
        if(listener != null) {
            listener.endSession(mailSourceStatus);
        }
    }
    
    /**
     * Add a new item from the server to the mail store.
     */
    public int addItem(SyncItem item) throws SyncException {
        Log.info("New item " + item.getKey() + " from server.");
        
        if(item.getType().equals(EMAIL_TYPE)) {
            // Parse the mail item
            Log.debug("Email item");
            return handleEmailItem(item);
            
        } else if(item.getType().equals(FOLDER_TYPE)) {
            // Parse the mail item
            Log.debug("Folder item");
            return addFolderItem(item);
            
        } else {
            Log.error("[MailSyncSource.addItem] Unknown item type: "
                    + item.getType());
            return 500;
        }
    }
    
    /** Update a given SyncItem stored on the source backend */
    public int updateItem(SyncItem item) throws SyncException {
        Log.info("Updated item " + item.getKey() + " from server.");
        if(item.getType().equals(EMAIL_TYPE)) {
            // Parse the mail item
            Log.debug("Email item");
            return handleEmailItem(item);
        } else if(item.getType().equals(FOLDER_TYPE)) {
            // Parse the mail item
            Log.debug("Folder item update: ignored");
            return 200;

        } else {
            Log.info("[MailSyncSource.updateItem] Unexpected item type"
                    + item.getType());
            return 500;
        }
    }
    
    /** Delete a SyncItem stored on the related Items list */
    public int deleteItem(String key) throws SyncException {
        Log.info("Delete from server for item " + key);
        try {
            Log.debug("[deleteItem]Email item");
            
            String parent = folderFromKey(key);
            if(parent == null){
                Log.error("Invalid key from server: " +key);
                return 500;
            }
            
            Folder f = store.getFolder(parent);
            String msgId = msgIdFromKey(key);
            f.deleteMessage(msgId);
            
            if(listener != null) {
                listener.msgDeleted(msgId);
            }                
            
            return 200;
            
        } catch (MailException me) {
            Log.error("Error removing message: " + key
                    + " " + me.toString() );
            mailSourceStatus = AccountListener.STATUS_RECV_ERROR;
            return 500;
        }
    }
    
    /**
     * Tell the SyncSource the status returned by the server
     * for an Item previously sent.
     *
     * @param key the key of the item
     * @param status the status code received for that item
     *
     * @throws SyncException if the SyncSource wants to stop the sync
     */
    public void setItemStatus(String key, int status)
    throws SyncException {
        Log.info("Status " + status + " for item " + key + " from server.");
        try {
            String folder = folderFromKey(key);
            String msgId = msgIdFromKey(key);
            
            if(Store.OUTBOX.equals(folder)) {
                if (SyncMLStatus.isSuccess(status)) {
                    
                    // Move the message in the sent folder
                    Folder sent = store.getFolder(Store.SENT);
                    Folder outbox = store.getFolder(Store.OUTBOX);
                    
                    Message msg = outbox.getMessage(msgId);
                    
                    msg.getFlags().setFlag(MessageFlags.TX_SENDING, false);
                    msg.getFlags().setFlag(MessageFlags.TX_SENT, true);
                    sent.appendMessage(msg);
                    outbox.deleteMessage(msg);
                    
                    if(listener != null) {
                        listener.msgSent();
                    }
                } else {
                    mailSourceStatus = AccountListener.STATUS_SEND_ERROR;
                    // TODO: send an event to the client?
                }
            } 
            
            if (Store.INBOX.equals(folder)) {
                MessageFlags newMask = FlagQueue.get(key);
                if (newMask!=null
                    &&(newMask.isSet(MessageFlags.ANSWERED)
                    ||newMask.isSet(MessageFlags.FORWARDED))) {
                    Folder inbox = store.getFolder(Store.INBOX);
                    Message msg = inbox.getMessage(msgId);
                    msg.setFlags(newMask);
                    inbox.appendMessage(msg);
                    listener.msgFlagsUpdated(msg.getMessageId(), msg.getFlags());
                }
            }
            
            FlagQueue.remove(key);
            
        } catch (MailException me) {
            me.printStackTrace();
            Log.error("setItemStatus: error setting the status for item:"+key);
        }
    }
    
    /**
     * Extends BaseSyncSource.getNextItem() to send events
     */
    public SyncItem getNextItem() throws SyncException {
        SyncItem ret = super.getNextItem();
        
        // If it's the last item, send the endSending() to the listener
        if(ret == null && listener != null) {
            listener.endSending();
        }
        
        return ret;
    }
    
    /**
     * Extends BaseSyncSource.getNextNewItem() to send events
     */
    public SyncItem getNextNewItem() throws SyncException {
        SyncItem ret = super.getNextNewItem();
        
        // If it's the last item, send the endSending() to the listener
        if(ret == null && listener != null) {
            listener.endSending();
        }
        
        return ret;
    }
    
    /**
     * Extends BaseSyncSource.dataReceived() to send events
     */
    public void dataReceived(String date, int size) {
        if(date != null && listener != null) {
            listener.dataReceived(date, size);
        }
    }

    /**
     * Send incomingMsgNumber event when the engine announce the number of
     * changes: it is the first Sync tag from server.
     */
    public void setServerItemsNumber(int number) {
        super.setServerItemsNumber(number);
        
        if (!receiving && listener != null) {
            listener.startReceiving(number);
        }
        receiving = true;
    }
    
    //------------------------------------------------------- Protected methods
    
    /**
     * Init the allItems list used in the slow sync with the messages
     * in the outbox only.
     */
    protected void initAllItems() throws SyncException {
        
        
        if (listener != null) {
            listener.startSending(0);
        }
        
    /*
        FIXME: check with the server why the folders are not accepted.

        allItems = new SyncItem[2];
        FolderData inbox = new FolderData("Inbox", new Date(), "inbox");
        FolderData outbox = new FolderData("Outbox", new Date(), "outbox");
     
        allItems[0] = new SyncItem ("Inbox",
                FOLDER_TYPE,
                SyncItem.STATE_UPDATED,
                "/",
                inbox.format());
     
        allItems[1] = new SyncItem ("Outbox",
                FOLDER_TYPE,
                SyncItem.STATE_UPDATED,
                "/",
                outbox.format());
     
     */
        
     /*
        try {
            // Get the keys of the messages in the Outbox only.
            Folder outbox = store.getFolder(Store.OUTBOX);
            Message[] headers = outbox.getMsgHeaders();
            int count = headers.length;
      
            if (count > 0) {
                allItems = new SyncItem[count];
      
                for(int i=0; i<count; i++) {
                    allItems[i] = new SyncItem(headers[i].getMessageId());
                    allItems[i].setParent(Store.OUTBOX);
                    allItems[i].setState(SyncItem.STATE_UPDATED);
                }
            }
            else {
                allItems = null;
            }
        }
        catch (MailException me) {
            Log.error("[MailSyncSource.initAllItems]: " + me.toString());
            throw new SyncException(me.toString());
        }
      */
        
    }
    
    /**
     * Init the newItems list used in the fast sync with the messages
     * in the outbox only.
     * TODO: check the state of the message
     */
    protected void initNewItems() throws SyncException {
        try {
            // Get the keys of the messages in the Outbox only.
            Folder outbox = store.getFolder(Store.OUTBOX);
            Message[] headers = outbox.getMsgHeaders();
            int count = headers.length;
            
            if (count > 0) {
                newItems = new SyncItem[count];
                
                for(int i=0; i<count; i++) {
                    String key = makeItemKey(Store.OUTBOX,
                            headers[i].getMessageId());
                    newItems[i] =
                            new SyncItem(key, EMAIL_TYPE,
                            SyncItem.STATE_NEW, Store.OUTBOX, null);
                }
            } else {
                newItems = null;
            }
            
            if (listener != null) {
                Log.debug(" init new items start sending call. Count =  " + count);
                listener.startSending(count);
            }
            
        } catch(MailException me) {
            Log.error("[MailSyncSource.initNewItems]: " + me.toString());
            me.printStackTrace();
            throw new SyncException(SyncException.CLIENT_ERROR, me.toString());
        }
    }
    
    /**
     * Not used now. It will be implemented if the Draft folder
     * has to be synchronized.
     */
    protected void initUpdItems() throws SyncException {
        Log.debug("Initializating Updated Items of folder INBOX");
        int count = 0;
        updItems = null;
        
        Hashtable updatedItemsList = FlagQueue.getUpdatedItemsList();
        if (updatedItemsList==null) {
            return;
        }
        count = updatedItemsList.size();
        
        if (count > 0) {
            int i=0;

            updItems = new SyncItem[count];
            Enumeration itemKeys = updatedItemsList.keys();
            while (itemKeys.hasMoreElements()) {
                String key = (String) itemKeys.nextElement();
                updItems[i] =
                        new SyncItem(key, EMAIL_TYPE,
                            SyncItem.STATE_UPDATED, folderFromKey(key), null);
                i++;
            }
        }
    }
    
    /**
     * Init the delItems list used in the fast sync
     * TODO: check the state of the message
     */
    protected void initDelItems() throws SyncException {
        Log.debug("Initializating Deleted Items");
        int count=0;
        delItems = null;
        Hashtable deletedItemsList = FlagQueue.getDeletedItemsList();
            
        if (deletedItemsList==null) {
            return;
        }
        count = deletedItemsList.size();
        if (count > 0) {
            delItems = new SyncItem[count];
            Enumeration itemKeys = deletedItemsList.keys();
            int i=0;
            while (itemKeys.hasMoreElements()) {
                
                String key = (String) itemKeys.nextElement();
                Log.debug("Folder to find: " + folderFromKey(key));
                delItems[i] =
                        new SyncItem(key, EMAIL_TYPE,
                        SyncItem.STATE_DELETED, folderFromKey(key), null);
                i++;
            }
        }
        
        /*
         *we don't need to notify the listener here
         if (listener != null) {
           listener.startSending(count);
        }*/
    }
    
    /**
     * This function gets the item content in the backend database and
     * returns a complete item. The parameter item is marked final because
     * should not be used for the filled item: it is a reference to the
     * array entry, and filling it would cause the array to keep all the
     * filled items at the end (the gc will not dispose them). <p>
     * The content of the item depends also from the encoding of this
     * SyncSource:
     * <li> if the encoding is <i>none</i>, it must be a String, converted
     *      with getBytes(), so the engine will send it unchanged.
     * <li> if the encoding is <i>b64</i>, the content can be binary, and the
     *      type should be set accordingly, so that the receiving source
     *      can handle it. In this way, the binary content is transferred
     *      encoded in the SyncML message. This encoding can be applied to
     *      a text item too, to avoid problems with charset or other, like
     *      what is done with the SIF format.
     */
    protected SyncItem getItemContent(final SyncItem item) throws SyncException {
        Log.debug("Current Item State: " + item.getState());
        try {
            EmailData mdata = null;
            SyncItem ret = null;
            switch(item.getState()){
                case SyncItem.STATE_UPDATED:
                    Folder folder = store.getFolder(item.getParent());
                    if (folder.getName().equals(Store.INBOX)) {
                        //Last result to be achieved:
                        mdata = new EmailData();
                        mdata.setFlags(FlagQueue.get(item.getKey()));
                        ret = new SyncItem(item);
                        ret.setContent(mdata.format());
                        Log.debug(new String(mdata.format()));
                        return ret;
                    }
                    break;
                    
                case SyncItem.STATE_NEW:
                    if(item.getType().equals(EMAIL_TYPE)) {
                        // For type email, the content has to be filled,
                        // for folders, the item is already complete.
                        // for deleted items
                        Folder f = store.getFolder(item.getParent());
                        if (f.getName().equals(Store.OUTBOX)) {
                            // Remove the folder prefix from the key.
                            String messageId = msgIdFromKey(item.getKey());
                            // Get the message
                            Message msg = f.getMessage(messageId);
                            mdata = new EmailData(msg);
                            ret = new SyncItem(item);
                            ret.setContent(mdata.format());
                            Log.debug("getItemContent: returning filled item.");
                            
                            // Return the filled item
                            return ret;
                        }
                    }
                    break;
                default:
                    break;
            }
            
        } catch (MailException me) {
            Log.error("[MailSyncSource.getItemContent] Error filling item:"
                    + item.getKey());
            me.printStackTrace();
            throw new SyncException(SyncException.CLIENT_ERROR, me.toString());
        }

        // If no changes, return the original item
        Log.debug("getItemContent: returning original item.");
        return item;
    }
    
    //-------------------------------------------------------- Private methods
    
    /**
     * Add an Email item
     */
    private int handleEmailItem(SyncItem item) {
        
        // Get parent folder name (handle 3.0 compatibility)
        String parent = folderFromParent(item.getParent());
        if(parent == null) {
            Log.error("[handleEmailItem]Invalid parent from server: "
                    +item.getParent());
            return 500;
        }
        // Get only Inbox messages, ignore the others
        if(!parent.equals(store.INBOX)) {
            Log.debug("[handleEmailItem]Message for "+parent+" discarded");
            return 200;
        }
        
        EmailData mdata = new EmailData();
        
        try {
            mdata.parse(item.getContent());
        } catch (XmlException xe) {
            Log.error("[handleEmailItem]Error parsing item: " + item.getKey() +
                      " ("+xe.toString()+")");
            xe.printStackTrace();
            // Return error to the server
            return 500;
        } catch (MailException me) {
            Log.error("[handleEmailItem]Error parsing message: "+item.getKey() +
                      " ("+me.toString()+")");
            me.printStackTrace();
            // Return error to the server
            return 500;
        } catch (Exception e) {
            Log.error("[handleEmailItem]Unexpected exception parsing message: "
                        + item.getKey() 
                        + " ("+e.toString()+")");
            e.printStackTrace();
            // Return error to the server
            return 500;
        }
        
        int ret = 500;

        switch(item.getState()) {
            case SyncItem.STATE_NEW:
                String localkey = addMessage(parent, mdata.getEmailItem());
                if (localkey != null) {
                    item.setKey(localkey);
                    
                    if(listener != null) {
                        listener.msgReceived(mdata.getEmailItem());
                    }
                    ret = 200;
                }
                break;
            case SyncItem.STATE_UPDATED:
                ret = updateFlags(parent, item.getKey(), mdata.getFlags());
                
                if(ret == 200 && listener != null) {
                    listener.msgFlagsUpdated(msgIdFromKey(item.getKey()), mdata.getFlags());
                }
                
                break;
            default:
                Log.error("[handleEmailItem]Unknown state: "+item.getState());
                break;
        }
        
        return ret;
    }

    /**
     * Add a message.
     */
    private String addMessage(String parent, Message msg) {
        String localkey = null;

        if (msg != null) {
            String msgid = msg.getMessageId();
            try {
                Folder f = store.getFolder(parent);
                Log.info("[addMessage]Adding message " + msgid +
                        " to folder " + f.getName());
                f.appendMessage(msg);
                
                // Set item key to LUID
                localkey = makeItemKey(f.getName(), msg.getMessageId());
                
            } catch (MailException me) {
                Log.error("[addMessage]Error saving message: " + msgid 
                            + " " + me.toString() );
                me.printStackTrace();
                // TODO: implement a finer handling of the exception?
                mailSourceStatus = AccountListener.STATUS_RECV_ERROR;
            }
        } else {
            // Server can't send flags update only in a Add command.
            Log.debug("[addMessage]Invalid empty email data: ");
            mailSourceStatus = AccountListener.STATUS_SERVER_ERROR;
        }
        return localkey;
    }
    
    /**
     * Update flags.
     */
    private int updateFlags(String parent, String key, MessageFlags flags) {
        try {
            Folder f = store.getFolder(parent);

            // Flag update only.
            Log.debug("[updateMessage]updating flags for: ", key);

            String msgId = msgIdFromKey(key);
            Message saved = f.getMessage(msgId);

            saved.setFlags(flags);
            f.appendMessage(saved);

            return 200;

        } catch (MailException me) {

            Log.error("[addMessage]Error saving message: " + key 
                    + " " + me.toString() );
            me.printStackTrace();
            // TODO: implement a finer handling of the exception?
            mailSourceStatus = AccountListener.STATUS_RECV_ERROR;

            return 500;
        }
    }
    
    /**
     * Add an Folder item
     */
    private int addFolderItem(SyncItem item) {
        
        FolderData fdata = new FolderData();
        
        try {
            // For the moment, the client don't creates
            // other folders on server's request, just checks
            // the commands to return the correct status
            // and key
            fdata.parse(item.getContent());
            // Set item key to LUID
            item.setKey(fdata.getName());
            return 200;
            
        } catch (XmlException xe) {
            Log.error("Error parsing item: " + item.getKey());
            // Return error to the server
            return 500;
        }
    }
    
    /*
     * Convert Funambol 3.0 folder names into local folder names.
     *
     * @param parent the item parent string
     * @return the converted name, or the original one if does not
     *         starts with "ROOT/"
     *
     */
    private String folderFromParent(String parent) {
        final String root = "ROOT/";
        
        if(parent == null) {
            return null;
        }
        if (parent.startsWith(root)) {
            return folderFromPrefix(parent.charAt(root.length()));
        }
        return parent;
    }
    
    /*
     * Get the folder name from the key.
     */
    private String folderFromKey(String key) {
        if(key == null) {
            return null;
        }
        int pos = key.indexOf('/');
        
        if(pos != -1) {
            return key.substring(0, pos);
        }
        return null;
    }
    
    /*
     * Get the message Id from the key.
     */
    private String msgIdFromKey(String key) {
        int pos = key.indexOf('/');
        
        if(pos != -1) {
            return key.substring(pos+1);
        }
        return key;
    }
    
    /**
     * Find the folder name from the Funambol key prefix.
     */
    private String folderFromPrefix(char prefix) {
        switch (prefix) {
            case 'I':
                return "Inbox";
            case 'O':
                return "Outox";
            case 'D':
                return "Drafts";
            case 'S':
                return "Sent";
            default:
                return null;
        }
    }
    
    /**
     * Compose the item key adding the folder prefix to the local key.
     * If it's one of the main folders, it is shortened, otherwise
     * the whole folder name is used as prefix.
     */
    private String makeItemKey(String folder, String key) {
        return folder+"/"+key;
    }
    
}


